@@ -89,7 +89,7 @@
 #endif
 #include <limits.h>
 
-#include "cwlib.h"
+#include <cwlib.h>
 #include "cwdaemon.h"
 
 /* network vars */
@@ -113,6 +113,7 @@ int soundcard_sound = 0;	/* soundcard off */
 int forking = 1; 		/* we fork by default */
 int bandswitch;
 int priority = 0;
+int async_abort = 0;
 
 /* flags for different states */
 int ptt_timer_running = 0;	/* flag for PTT state */
@@ -139,14 +140,14 @@ cwdevice cwdevice_ttys = {
 };
 
 cwdevice cwdevice_null = {
-       init:null_init,
-       free:null_free,
-       reset:null_reset,
-       cw:null_cw,
-       ptt:null_ptt,
-       ssbway:NULL,
-       switchband:NULL,
-       footswitch:NULL
+   init:null_init,
+   free:null_free,
+   reset:null_reset,
+   cw:null_cw,
+   ptt:null_ptt,
+   ssbway:NULL,
+   switchband:NULL,
+   footswitch:NULL
 };
 
 #if defined (HAVE_LINUX_PPDEV_H) || defined (HAVE_DEV_PPBUS_PPI_H)
@@ -162,6 +163,7 @@ cwdevice cwdevice_lp = {
 };
 #endif
 cwdevice *cwdev;
+static void playmorsestring (char *x);
 
 /* catch ^C when running in foreground */
 static RETSIGTYPE
@@ -350,8 +352,12 @@ recv_code (void)
 		message[recv_rc] = '\0';
 		if (message[0] != 27)
 		{	/* no ESCAPE */
+			debug ("Message = %s", message);
 			if ((strlen (message) + strlen (morsetext)) <= MAXMORSE - 1)
+			{
 				strcat (morsetext, message);
+				playmorsestring (morsetext);
+			}
 			return 1;
 		}
 		else
@@ -367,6 +373,7 @@ recv_code (void)
 				soundcard_sound = 0;
 				initmorse ();
 				wordmode = 0;
+				async_abort = 0;
 				cwdev->reset (cwdev);
 				debug ("Reset all values");
 				break;
@@ -401,11 +408,10 @@ recv_code (void)
 				break;
 			case '4':	/* message abort */
 				debug ("Message abort");
-				if (sendingmorse == 1)
-				{
-					cw_flush_tone_queue ();
-					aborting = 1;
-				}
+				strcpy (morsetext, "");
+				cw_flush_tone_queue ();
+				cw_wait_for_tone_queue ();
+				aborting = 1;
 				break;
 			case '5':	/* exit */
 				cwdev->free (cwdev);
@@ -424,12 +430,13 @@ recv_code (void)
 					break;
 				if ((lv > -51) && (lv < 51))	/* only allowed range */
 				{	
-					cw_set_weighting (lv);
+					cw_set_weighting (lv * 0.6 + 50);
 					debug ("Weight: %ld", lv);
 				}
 				break;
 			case '8':	/* device type */
 				ndev = message + 2;
+				debug ("Device: %s", ndev);
 				if ((fd = dev_is_tty(ndev)) != -1)
 				{
 					cwdev->free(cwdev);
@@ -565,14 +572,15 @@ recv_code (void)
 					cw_set_volume (morse_volume);
 				}
 				break;
-	    case 'h':   /* send echo to main program when CW playing is done */
-		memcpy(&reply_sin, &k_sin, sizeof(reply_sin)); /* remember sender */
-		reply_socklen = sin_len;
-		strncpy(reply_data+1, message, sizeof(reply_data) - 2);
-		reply_data[sizeof(reply_data) - 1] = '\0';
-		reply_data[0]='h';
-		if (strlen (message) + 1 <= MAXMORSE - 1) strcat (morsetext, "^");
-		break;
+			case 'h':   /* send echo to main program when CW playing is done */
+				memcpy (&reply_sin, &k_sin, sizeof(reply_sin)); /* remember sender */
+				reply_socklen = sin_len;
+				strncpy (reply_data + 1, message, sizeof(reply_data) - 2);
+				reply_data[sizeof(reply_data) - 1] = '\0';
+				reply_data[0]='h';
+				if (strlen (message) + 1 <= MAXMORSE - 1)
+					strcat (morsetext, "^");
+				break;
 			}
 			return 0;
 		}
@@ -590,11 +598,6 @@ playmorsestring (char *x)
 	sendingmorse = 1;
 	while (*x)
 	{
-		if (aborting == 1)
-		{
-			aborting = 0;
-			break;
-		}
 		c = *x;
 		if ((c == '+') || (c == '-'))
 		{		/* speed in- & decrease */
@@ -606,15 +609,15 @@ playmorsestring (char *x)
 		}
 		else if (c == '~')
 			cw_set_gap (2); /* 2 dots time additional for the next char */
-	else if (c == '^'){
-	    debug("Echo '%s'", reply_data);
-	    if (strlen(reply_data)==0) return;
-	    sendto(socket_descriptor, reply_data, strlen(reply_data), 0, 
-		(struct sockaddr *)&reply_sin, reply_socklen);
-	    strcpy(reply_data,"");
-	    break;
-  
-	}
+		else if (c == '^')
+		{
+			debug ("Echo '%s'", reply_data);
+			if (strlen (reply_data) == 0) return;
+			sendto (socket_descriptor, reply_data, strlen (reply_data), 0, 
+				(struct sockaddr *)&reply_sin, reply_socklen);
+			strcpy (reply_data,"");
+			break;
+		}
 		else
 		{
 			validchar = 1;
@@ -633,9 +636,9 @@ playmorsestring (char *x)
 					udelay (ptt_delay);
 				}
 			}
+			if (c == '*') c = '+';
 			debug ("Morse = %c", c);
 			cw_send_character (c);
-			cw_tone_queue_wait_critical (1);
 			if (cw_get_gap () == 2) cw_set_gap (0);
 		}
 		x++;
@@ -660,7 +663,7 @@ playmorsestring (char *x)
 	sendingmorse = 0;
 }
 
-static void keyingevent (int keystate)
+static void keyingevent (void *arg, int keystate)
 {
 	if (keystate == 1)
 		cwdev->cw (cwdev, ON);
@@ -668,14 +671,6 @@ static void keyingevent (int keystate)
 		cwdev->cw (cwdev, OFF);
 }
 
-static void pttevent (int pttstate)
-{
-	if (pttstate == 1)
-		cwdev->ptt(cwdev,ON);
-	else
-		cwdev->ptt(cwdev,OFF);
-}
-
 /* parse the command line and check for options, do some error checking */
 static void
 parsecommandline (int argc, char *argv[])
@@ -824,8 +819,7 @@ int
 main (int argc, char *argv[])
 {
 	pid_t pid, sid;
-	int fd;
-	int save_file_flags;
+	int fd, fd_count, save_flags;
 
 #if defined HAVE_LINUX_PPDEV_H
 	keydev = "parport0";
@@ -865,8 +859,7 @@ main (int argc, char *argv[])
 	cwdev->desc = keydev;
 
 	initmorse ();
-	cw_keying_callback (keyingevent);
-	cw_ptt_callback (pttevent);
+	cw_register_keying_callback (keyingevent, NULL);
 
 	debug ("Device used: %s", cwdev->desc);
 	cwdev->init (cwdev, fd);
@@ -900,28 +893,26 @@ main (int argc, char *argv[])
 		/* replace stdin/stdout/stderr with /dev/null */
 		if ((fd = open("/dev/null", O_RDWR, 0)) == -1)
 		{
-		        syslog (LOG_ERR, "%s\n", "open null");
-		        exit (1);
+			syslog (LOG_ERR, "%s\n", "open null");
+			exit (1);
 		}
-		if (dup2(fd, STDIN_FILENO) == -1)
+		if (dup2 (fd, STDIN_FILENO) == -1)
 		{
-		        syslog (LOG_ERR, "%s\n", "dup2 stdin");
-		        exit (1);
+			syslog (LOG_ERR, "%s\n", "dup2 stdin");
+			exit (1);
 		}
-		if (dup2(fd, STDOUT_FILENO) == -1)
+		if (dup2 (fd, STDOUT_FILENO) == -1)
 		{
-		        syslog (LOG_ERR, "%s\n", "dup2 stdout");
-		        exit (1);
+			syslog (LOG_ERR, "%s\n", "dup2 stdout");
+			exit (1);
 		}
-		if (dup2(fd, STDERR_FILENO) == -1)
+		if (dup2 (fd, STDERR_FILENO) == -1)
 		{
-		        syslog (LOG_ERR, "%s\n", "dup2 stderr");
-		        exit (1);
+			syslog (LOG_ERR, "%s\n", "dup2 stderr");
+			exit (1);
 		}
-		if (fd != STDIN_FILENO &&
-		    fd != STDOUT_FILENO &&
-		    fd != STDERR_FILENO)
-		        (void)close(fd);
+		if (fd != STDIN_FILENO && fd != STDOUT_FILENO && fd != STDERR_FILENO)
+			(void)close (fd);
 	}
 	else
 	{
@@ -949,15 +940,15 @@ main (int argc, char *argv[])
 		exit (1);
 	}
 
-	save_file_flags = fcntl (socket_descriptor, F_GETFL);
-	if (save_file_flags == -1)
+	save_flags = fcntl (socket_descriptor, F_GETFL);
+	if (save_flags == -1)
 	{
-		errmsg("Trying get flags");
-		exit(1);
+		errmsg ("Trying get flags");
+		exit (1);
 	}
-	save_file_flags |= O_NONBLOCK;
+	save_flags |= O_NONBLOCK;
 
-	if (fcntl (socket_descriptor, F_SETFL, save_file_flags) == -1)
+	if (fcntl (socket_descriptor, F_SETFL, save_flags) == -1)
 	{
 		errmsg ("Trying non-blocking");
 		exit (1);
@@ -975,13 +966,24 @@ main (int argc, char *argv[])
 #endif
 
 	morsetext[0] = '\0';
-	while (1)
+	do
 	{
-		udelay (500);	/*prevent 100% CPU */
+		fd_set readfd;
+		struct timeval udptime;
+
+		FD_ZERO (&readfd);
+		FD_SET (socket_descriptor, &readfd);
+		udptime.tv_sec = 0;
+		udptime.tv_usec = 500;
+		fd_count = select (socket_descriptor + 1, &readfd, NULL, NULL, &udptime);
+		if (fd_count == -1 && errno != EINTR)
+			errmsg ("Select");
+		else
+			recv_code ();
+
 		if (cwdev->footswitch)
-			cw_ptt_control (!((cwdev->footswitch(cwdev))));
-		if (recv_code ())
-			playmorsestring (morsetext);
+			cwdev->ptt (cwdev, !((cwdev->footswitch(cwdev))));
+
 		/* check for ptt off timer */
 		if (1 == ptt_timer_running)
 		{
@@ -993,7 +995,7 @@ main (int argc, char *argv[])
 				ptt_timer_running = 0;
 			}
 		}
-	}
+	} while (1);
 
 	cwdev->free (cwdev);
 	if (close (socket_descriptor) == -1)
